use std::{
    collections::HashMap,
    io::{BufRead, BufReader, Write},
    net::{TcpListener, TcpStream},
};

const VALID_METHODS: [&str; 8] = [
    "GET", "HEAD", "POST", "PUT", "DELETE", "CONNECT", "OPTIONS", "TRACE",
];

#[derive(Debug)]
struct StartLine {
    method: String,
    target: String,
    version: String,
}

impl StartLine {
    pub fn new() -> Self {
        Self {
            method: String::new(),
            target: String::new(),
            version: String::new(),
        }
    }

    pub fn is_valid_method(method: &String) -> bool {
        if method.trim().is_empty() {
            return false;
        }
        let tmp = method.as_str();

        if VALID_METHODS.contains(&tmp) {
            return true;
        } else {
            return false;
        }
    }

    pub fn is_valid_target(target: &String) -> bool {
        if target.trim().is_empty() {
            return false;
        }

        return true;
    }

    pub fn is_valid_version(version: &String) -> bool {
        if version.trim().is_empty() {
            return false;
        }

        return true;
    }
}

fn parse_start_line(input: String) -> Option<StartLine> {
    let mut start_line = StartLine::new();
    let vec = input.trim().split(" ").collect::<Vec<&str>>();

    if vec.len() != 3 {
        return None;
    }

    let method = String::from(vec[0]);
    if StartLine::is_valid_method(&method) {
        start_line.method = method;
    }

    let target = String::from(vec[1]);
    if StartLine::is_valid_target(&target) {
        start_line.target = target;
    }

    let version = String::from(vec[2]);
    if StartLine::is_valid_version(&version) {
        start_line.version = version;
    }

    return Some(start_line);
}

fn handle_request(mut stream: TcpStream) {
    let mut line = String::new();
    let mut reader = BufReader::new(&mut stream);
    reader.read_line(&mut line).unwrap();

    let mut counter = 0;
    let max_preceding_empty_lines = 1000;

    // Request can have one or many empty lines preceding the start-line and I will ignore these
    // I will also for now only allow up to 1000 of these empty lines before I abort
    loop {
        if counter > max_preceding_empty_lines {
            stream
                .write_all(
                    b"HTTP/1.1 400 Bad Request\r\n\r\nReceived too many preceding empty lines",
                )
                .unwrap();
            return;
        }

        // I will not support ignoring the CR
        if !line.ends_with("\r\n") {
            stream
                .write_all(b"HTTP/1.1 400 Bad Request\r\n\r\nLines must end with CRLF")
                .unwrap();
            return;
        }
        line = line.trim().into();

        if !line.is_empty() {
            break;
        }
        reader.read_line(&mut line).unwrap();
        counter += 1;
    }

    let start_line = match parse_start_line(line) {
        Some(val) => val,
        None => {
            stream
                .write_all(b"HTTP/1.1 400 Bad Request\r\n\r\nInvalid start-line")
                .unwrap();
            return;
        }
    };

    dbg!(&start_line);

    if start_line.method != "GET" {
        stream
            .write_all(b"HTTP/1.1 501 Not Implemented\r\n\r\nServer currently only supports GET")
            .unwrap();
        return;
    }

    let mut is_first_line = true;
    let mut field_lines: HashMap<String, String> = HashMap::new();

    loop {
        let mut line = String::new();
        reader.read_line(&mut line).unwrap();
        // print!("{line}");

        if line.starts_with(" ") {
            if !is_first_line {
                continue;
            }
            stream
                .write_all(b"HTTP/1.1 400 Bad Request\r\n\r\nCannot have whitespace between the start-line and field-lines")
                .unwrap();
            return;
        }

        // I will not support ignoring the CR
        if !line.ends_with("\r\n") {
            stream
                .write_all(b"HTTP/1.1 400 Bad Request\r\n\r\nLines must end with CRLF")
                .unwrap();
            return;
        }

        if line.trim().is_empty() {
            break;
        }

        let field_line = line.split_once(":").expect("Shits fucked: {}");
        field_lines.insert(field_line.0.to_owned(), field_line.1.trim().to_owned());

        is_first_line = false;
    }

    dbg!(&field_lines);

    stream
        .write_all(b"HTTP/1.1 200 OK\r\n\r\nHello, World!\r\n")
        .unwrap();
}

fn main() {
    let listener = TcpListener::bind("127.0.0.1:8080").unwrap();

    for stream in listener.incoming() {
        let stream = stream.unwrap();
        handle_request(stream)
    }
}
